#encoding=utf8
import json
import os
import copy
import requests
import time
import logging
import shutil
import configparser
import uuid
import datetime
from os.path import join, dirname, abspath
from requests.packages.urllib3.exceptions import InsecureRequestWarning
from logging.handlers import RotatingFileHandler
from cshmac.requests import HmacAuth

requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

config = configparser.ConfigParser()
conf_path = join(dirname(abspath(__file__)), 'setup.conf')
config.read(conf_path)
MAIN_URL = 'https://' + config['DEFAULT']['MAIN_URL']
APPLIANCE_PATH = config['DEFAULT']['APPLIANCE_PATH']
LOG_DIR = config['DEFAULT']['LOG_DIR']
LOG_FILE_PATH = join(LOG_DIR, 'sample_data.log')

formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)
handler = RotatingFileHandler(LOG_FILE_PATH, maxBytes=5 * 1024 * 1024, backupCount=10)
handler.setFormatter(formatter)
logger.addHandler(handler)
logger.setLevel('INFO')

APPLIANCE_PUBLIC_KEY = ''
APPLIANCE_PRIVATE_KEY = ''
WORKFLOW_COLLECTIONS = '/api/3/workflow_collections'
STAGING_MODEL_MEATDATAS = '/api/3/staging_model_metadatas'
PUBLISH = '/api/publish'
SYSTEM_VIEW_TEMPLATES = '/api/3/system_view_templates'
REPORTS = '/api/3/template/reporting/'
ROLES = '/api/3/roles'
DASHBOARDS = '/api/3/template/dashboard/'
PAGE = '/api/3/system_view_templates?$export=true'
EMAIL_TEMPLATE = '/api/3/email_templates/'
WORK_FLOW_STEP = '/api/3/workflow_steps/'
TAGS = '/api/3/bulkupsert/tags'


def make_request(url, method, body=None):
    try:
        bodyless_methods = ['head', 'get']
        if method.lower() in bodyless_methods:
            body = None
        auth = HmacAuth(url, method, APPLIANCE_PUBLIC_KEY, APPLIANCE_PRIVATE_KEY, json.dumps(body))
        response = requests.request(method, url, auth=auth, json=body, verify=False)
        if response.status_code == 200 or response.status_code == 201 or response.status_code == 204:
            return response.json()
        elif response.status_code == 503:
            return response.json()
        elif response.status_code == 400:
            logger.info(response.json())
    except Exception as Err:
        logger.exception('Exception occurred: {}'.format(str(Err)))
        return False


def get_roles_uuid():
    roles_url = MAIN_URL + ROLES
    roles = make_request(roles_url, 'GET')
    role_uuid = []
    for i in roles["hydra:member"]:
        role_uuid.append(i["@id"].rsplit('/', 1)[-1])
    return role_uuid

def add_bulk_tags(data_path,item):
    page_url = MAIN_URL + TAGS
    with open(data_path + '/' + item,'r') as fp:
        body = json.load(fp)
    tags = make_request(page_url,'POST',body=body)

def add_page(data_path,item):
    page_url = MAIN_URL + PAGE
    with open(data_path + '/' + item, 'r') as fp:
        body = json.load(fp)
    pages = make_request(page_url, 'POST', body=body)


def post_reports(data_path, item, role_uuid):
    report_template_url = MAIN_URL + REPORTS.rstrip('/')
    get_reports = make_request(report_template_url, 'GET')
    report_name = []
    with open(data_path + '/' + item, 'r') as fp:
        contents = json.loads(fp.read())
        if get_reports == []:
            contents["roleIds"] = role_uuid
            report_url = MAIN_URL + REPORTS + str(uuid.uuid1())
            make_request(report_url, 'POST', body=contents)
        else:
            for report in get_reports:
                report_name.append(report["displayName"])
            if contents["displayName"] not in report_name:
                contents["roleIds"] = role_uuid
                report_url = MAIN_URL + REPORTS + str(uuid.uuid1())
                make_request(report_url, 'POST', body=contents)
                report_name.append(contents["displayName"])
                logger.info("The imported reports are: {}".format(report_name))


def import_roles(roles):
    roles_url = MAIN_URL + '/api/3/roles'
    roles_data = make_request(roles_url, 'GET')
    # Get Module related data, i.e, @id of module
    module_data = make_request(MAIN_URL + '/api/3/modules?$limit=65432', 'GET')
    roles_names = []
    for role in roles_data['hydra:member']:
        roles_names.append(role['name'])
    module_names = {}
    for i in module_data['hydra:member']:
        module_names.update({i['type']: i['@id']})
    for element in roles:
        name = element['name']
        if name not in roles_names:
            logger.info('Adding new role')
            url = MAIN_URL + '/api/3/roles'
            module_permission = element["modulePermissions"]
            for each_module in module_permission:
                fieldPermissions = each_module["fieldPermissions"]
                if len(fieldPermissions) > 0:
                    for field in fieldPermissions:
                        field.pop('@id')
                if each_module['module'] in module_names.keys():
                    each_module['module'] = {"@id": module_names.get(each_module['module'])}
            element.pop('@id')
            make_request(url, 'POST', body=element)
        elif '@id' in element:
            id = element['@id']
            role_id_url = MAIN_URL + id + '?$relationships=true'
            module_permission = element["modulePermissions"]
            for each_module in module_permission:
                if each_module['module'] in module_names.keys():
                    each_module['module'] = {"@id": module_names.get(each_module['module'])}
            make_request(role_id_url, 'PUT', body=element)
        else:
            logger.info('@id is missing for role: {}'.format(name))


def import_dashboards(contents, role_uuid):
    dashboard_template_url = MAIN_URL + DASHBOARDS.rstrip('/')
    get_dashboards = make_request(dashboard_template_url, 'GET')
    dashboard_name = []
    if get_dashboards == []:
        for single_dashboard in contents:
            single_dashboard['roleIds'] = role_uuid
            dashboard_url = MAIN_URL + DASHBOARDS + str(uuid.uuid1())
            make_request(dashboard_url, 'POST', body=single_dashboard)
    else:
        for dashboard in get_dashboards:
            dashboard_name.append(dashboard['displayName'])
        for single_dashboard in contents:
            if single_dashboard["displayName"] not in dashboard_name:
                single_dashboard['roleIds'] = role_uuid
                dashboard_url = MAIN_URL + DASHBOARDS + str(uuid.uuid1())
                make_request(dashboard_url, 'POST', body=single_dashboard)
                dashboard_name.append(single_dashboard["displayName"])
                logger.info("The imported dashboards are: {}".format(dashboard_name))


def import_views(views):
    for view in views:
        views_url = MAIN_URL + SYSTEM_VIEW_TEMPLATES + '/' + view.get('uuid')
        make_request(views_url, 'PUT', body=view)


def post_playbook(playbook_url, body):
    try:
        response = make_request(playbook_url, 'POST', body=body)
        if response:
            logger.info("Playbook import response: {}".format(response))
    except requests.exceptions.Timeout as Err:
        logger.exception('Timeout exception: {}'.format(str(Err)))
    except requests.exceptions.RequestException as Err:
        logger.exception(str(Err))
        return


def trigger_playbook(main_url, payload=None):
    response = make_request(main_url, 'POST', body=APPLIANCE_PUBLIC_KEY)
    logger.info('Response for trigger playbook: {}'.format(response))
    poll_url = MAIN_URL + '/api/wf/api/workflows/?task_id={task_id}&fields=status&parent__isnull=True'.format(task_id=response["task_id"])
    current_time = datetime.datetime.now()
    max_wait_time = current_time + datetime.timedelta(minutes=10)
    while datetime.datetime.now() < max_wait_time:
        ret = make_request(poll_url, 'GET')
        hydra_member = ret.get("hydra:member")
        if len(hydra_member) == 0:
            time.sleep(20)
        elif ('status' in hydra_member[0] and hydra_member[0]['status'] == 'active') or ('status' in hydra_member[0] and hydra_member[0]['status'] == 'incipient'):
            time.sleep(30)
        else:
            logger.info("Playbook execution completed: {}".format(hydra_member[0]))
            return


def get_playbook_data(data_path, item):
    with open(data_path + '/' + item, 'r', encoding='utf-8') as fp:
        contents = json.loads(fp.read())
        playbook_url = MAIN_URL + WORKFLOW_COLLECTIONS
        if 'data' in contents:
            data = contents["data"]
        else:
            data = contents["hydra:member"]
        for playbook in data:
            post_playbook(playbook_url, playbook)


def post_modules_in_staging_model_metadata(modules):
    modules_imported = []
    api_url = MAIN_URL + STAGING_MODEL_MEATDATAS + "?$limit=100"
    # GET existing model_metadatas
    mmd_response = make_request(api_url, 'GET')
    if mmd_response:
        mmd = mmd_response["hydra:member"]
        for module in modules:
            response = make_request(api_url, 'POST', body=module)
            if response:
                logger.info("Module created")
                modules_imported.append(response)
            else:
                logger.info("Module already exists. Updating")
                module_type = module["type"]
                # GET module id
                module_id = next(item for item in mmd if item["type"] == module_type)["@id"]
                # PUT on module id
                modules_copy = copy.copy(module)
                module['attributes']=[]
                module['uniqueConstraint']=[]
                response = make_request(MAIN_URL + module_id, 'PUT', body=module)
                if response:
                    make_request(MAIN_URL + module_id, 'PUT', body=modules_copy)

                logger.info("module {} updated".format(module_type))
    return modules_imported


def publish_module():
    api_call = MAIN_URL + PUBLISH
    make_request(api_call, 'PUT', body={})
    logger.info("Publish started")
    while True:
        ret = make_request(MAIN_URL + '/api/publish/error', 'GET')
        if 'status' in ret and ret['status'] == 'Success':
            logger.info("Publish successful")
            return True
        elif 'status' in ret and ret['status'] == 'Failure':
            logger.info("Publish Failed")
            return False
        else:
            time.sleep(10)


# The import picklists logic is for CyOPs version 4.11.1. The chanegs for 4.11.0 are not yet added,because the configuration export file has @id in it.
def post_picklists(picklists):
    final_list = []
    for element in picklists:
        d1 = {}
        name = element['name']
        if ' ' in name:
            name = name.replace(" ", "+")
        picklist_names_url = MAIN_URL + '/api/3/picklist_names?name=' + name
        picklist_names = make_request(picklist_names_url, 'GET')
        if 'hydra:totalItems' in picklist_names and picklist_names['hydra:totalItems'] == 0:
            logger.info('Adding new picklist')
            url = MAIN_URL + '/api/3/'
            result = make_request(url, 'POST', body=[element])
            if result:
                d1 = {name: []}
                for x in element['picklists']:
                    picklist_value = x['itemValue']
                    d1[name].append(picklist_value)

        elif 'uuid' in element:
            id = element['uuid']
            picklist_url = MAIN_URL + '/api/3/picklist_names/' + id + '?$relationships=true'
            result = make_request(picklist_url, 'GET')
            name_picklist = []
            for k in result['picklists']:
                name_picklist.append(k['itemValue'])
            name_picklist = set(name_picklist)
            found_new_value = False
            for j in element['picklists']:
                picklist_value_name = j['itemValue']
                if picklist_value_name not in name_picklist:
                    found_new_value = True
                    result['picklists'].append(j)
                    if not d1:
                        d1 = {name: []}
                    d1[name].append(picklist_value_name)
            if found_new_value:
                result_ret = make_request(MAIN_URL + '/api/3/picklist_names/' + id, 'PUT', body=result)
                if not result_ret:
                    logger.error("Error occurred while updating the the picklist value for picklist: %s"%name)
        else:
            logger.info('UUID is missing for picklist: {}'.format(name))
        if d1:
            final_list.append(d1)
    if final_list:
        logger.info('Updated or New picklist data: {}'.format(final_list))


def get_modelmetadata_svt_from_file(filename):
    with open(filename) as fp:
        data = fp.read()
        contents = json.loads(data)
        modules = contents["modules"]
        viewTemplates = contents["viewTemplates"]
        picklists = contents["picklistNames"]
        dashboards = contents["dashboards"]
        views = contents["views"]
        roles = contents["roles"]
        post_picklists(picklists)
        modules_imported = post_modules_in_staging_model_metadata(modules)
        for key, val in viewTemplates.items():
            for element in val:
                # Add in SVT
                request = make_request(MAIN_URL + SYSTEM_VIEW_TEMPLATES + '/' + element.get('uuid'), 'PUT', body=element)
                if not request:
                    logger.info("New Module View Template detected, inserting same")
                    make_request(MAIN_URL + SYSTEM_VIEW_TEMPLATES, 'POST', body=element)
        return modules_imported, views, roles, dashboards


def read_files(key_path):
    global APPLIANCE_PUBLIC_KEY
    global APPLIANCE_PRIVATE_KEY
    try:
        with open(join(key_path, 'APPLIANCE_PRIVATE_KEY'), 'r') as private:
            APPLIANCE_PRIVATE_KEY = private.read().strip()
        with open(join(key_path, 'APPLIANCE_PUBLIC_KEY'), 'r') as public:
            APPLIANCE_PUBLIC_KEY = public.read().strip()
    except FileNotFoundError:
        src = APPLIANCE_PATH
        shutil.copy(join(src, 'APPLIANCE_PRIVATE_KEY'), key_path)
        shutil.copy(join(src, 'APPLIANCE_PUBLIC_KEY'), key_path)
        read_files(key_path)


def get_appliance_keys():
    key_path = join(os.path.dirname(os.path.abspath(__file__)), 'keys')
    read_files(key_path)


def modify_email_template(data_path):
    with open(data_path+'/email_template.json', 'r', encoding='utf-8') as fp:
        data = fp.read()
        contents = json.loads(data)
        for item in contents:
            request = make_request(
                MAIN_URL + EMAIL_TEMPLATE + item.get('uuid'), 'PUT', body=item)
            if not request:
                logger.info("Failed To Update Email Template")

    return


def modify_task_email_template(data_path):
    with open(data_path+'/task_playbook.json', 'rb') as fp:
        data = fp.read()
        contents = json.loads(data)
        for item in contents:
            request = make_request(
                MAIN_URL +WORK_FLOW_STEP + item.get('uuid'), 'PUT', body=item)
            if not request:
                logger.info("Failed To update send Task Email step")
                
    return


def main():
    get_appliance_keys()
    data_path = join(dirname(os.path.dirname(os.path.abspath(__file__))), 'data')
    role_uuid = get_roles_uuid()
    modules_imported = []
    modify_email_template(data_path)
    modify_task_email_template(data_path)
    for item in sorted(os.listdir(data_path)):
        logger.info("The file is: {}".format(item))
        if item.startswith('Config - 01 - MMD and SVT'):
            modules_imported, views, roles, dashboards = get_modelmetadata_svt_from_file(data_path + '/' + item)
        if item.startswith('Config - 03 - Picklists.json'):
            filename = data_path + '/' + item
            with open(filename) as fp:
                data = fp.read()
                contents = json.loads(data)
                picklists = contents["picklistNames"]
                post_picklists(picklists)
        if item.startswith('Report'):
            post_reports(data_path, item, role_uuid)
        if item.startswith('CustomTags'):
            add_bulk_tags(data_path,item)
        if item.startswith('Playbooks'):
            get_playbook_data(data_path, item)
        if item.startswith('Pages'):
            add_page(data_path,item)
    res = publish_module()
    if res:
        import_roles(roles)
        import_dashboards(dashboards, role_uuid)
        import_views(views)
        endpoint = 'demo_records'
        url = MAIN_URL + '/api/triggers/1/' + endpoint
        trigger_playbook(url)


if __name__ == "__main__":
    main()
